%function [res] = logdetG4(v,sig,co)
% Computes the logarithm determinant of large matrix using a dynamic
% coding of Cholesky decomposition. This avoids overflow/underflow problems
% that are likely to happen when applying to det to large matrices. The
% idea is based on the fact that the det of a triangular matrix equals the
% product of its diagonal elements. Therefore, the matrix log-det is equal
% to the sum of their log values
%
% Computes log(det(X)) where X is a guassian matrix of v with width sigma
%
% Input: 
%    v - large vector
%    sig - gaussian width parameter
%    co - number of temp files to use
%
% Output:
%    res - the log det of the Gaussian matrix
%
% David R. Hardoon 3/1/2009
% D.Hardoon@cs.ucl.ac.uk
%
% 4/1/2009 - Optimised speed by removing 2 for loops

%% set up
PATH = 'workdir/';
if ~isdir(PATH)
   mkdir(PATH);
end

m = size(v,2);

fm = ceil(m/co);
tmp = m;

for i=1:fm    
    if co <= tmp
        dyn = cell(co,1);
    else
        dyn = cell(tmp,1);
    end
    tmp = tmp - co;
        
    save([PATH 'D' num2str(i) '.mat'],'dyn');
end

%% First element
% gaussian
j = 1;
disp(['Working on element ' num2str(j)]);
[X2,Y2]=meshgrid(sum(v.^2),sum(v(:,j).^2));
fa = exp(-(X2'+Y2'-2*v'*v(:,j))/(2*mpower(sig,2)));

% cholesky
load([PATH 'D' num2str(1) '.mat']);
dyn{j} = sqrt(fa(j));
savedata = dyn{j}(end);

i=j+1:length(dyn);
intval = fa(i);
asgval = intval/dyn{j}(end);
dyn(i) = num2cell(vertcat([dyn{i}],asgval')',2);
save([PATH 'D' num2str(1) '.mat'],'dyn');

for k=2:fm
    load([PATH 'D' num2str(k) '.mat']);
    tm = length(dyn);
    i=1:tm;    
    intval = fa(i+(k-1)*co);
    asgval = intval/savedata;
    dyn(i) = num2cell(vertcat([dyn{i}],asgval')',2);
    save([PATH 'D' num2str(k) '.mat'],'dyn');
end

    
%% Remaining elements
% For each i going from 1 to n, do the following:
% 1. Subtract from X(i,i) the dot product of the ith row of L with itself
%    set the entry L(i,i) to the square root of this result
% 2. For each value of j going from i+1 n, do the following:
%    Substract from X(j,i) the dot product of rows i and j and set the
%    entry of L(j,i) to this results divided by L(i,i).
%
for j=2:m-1  
    disp(['Working on element ' num2str(j)]);
    % gaussian
    [X2,Y2]=meshgrid(sum(v.^2),sum(v(:,j).^2));
    fa = exp(-(X2'+Y2'-2*v'*v(:,j))/(2*mpower(sig,2)));
    
    % cholesky
    k = ceil(j/co);
    load([PATH 'D' num2str(k) '.mat']);
    
    jj = mod(j,co);
    if jj == 0
        jj = j/k;
    end
    
    val = dyn{jj}*dyn{jj}';
    dyn{jj} = [dyn{jj} sqrt(fa(j) - val)];    
    savedata = dyn{jj}(end);
    savevec = dyn{jj}(1:end-1)';
    
    i=jj+1:length(dyn);
    try
        intval = fa(i) - cell2mat(dyn(i))*savevec;
        asgval = intval/dyn{jj}(end);
        dyn(i) = num2cell([cell2mat(dyn(i)),asgval],2);  %% most expensive line as requires setting up the triangular matrix
        %dyn{j-1} = dyn{j-1}(end);
    catch
        % no worries
    end

    save([PATH 'D' num2str(k) '.mat'],'dyn');

    for k=k+1:fm
        load([PATH 'D' num2str(k) '.mat']);
        tm = length(dyn);
        i=1:tm;
        intval = fa(i+(k-1)*co) - cell2mat(dyn(i))*savevec;
        asgval = intval/savedata;
        dyn(i) = num2cell([cell2mat(dyn(i)),asgval],2);
        save([PATH 'D' num2str(k) '.mat'],'dyn');
    end
end

%% Last element
% gaussian
j = m;
disp(['Working on element ' num2str(j)]);
[X2,Y2]=meshgrid(sum(v.^2),sum(v(:,j).^2));
fa = exp(-(X2'+Y2'-2*v'*v(:,j))/(2*mpower(sig,2)));

k = ceil(j/co);
load([PATH 'D' num2str(k) '.mat']);

jj = mod(j,co);
if jj == 0
    jj = (j/k);
end

% cholesky
val = dyn{jj}*dyn{jj}';
dyn{jj} = [dyn{jj} sqrt(fa(j) - val)];
save([PATH 'D' num2str(k) '.mat'],'dyn');

featend = cellfun(@(x) x(end), dyn);
%dyn{j-1} = dyn{j-1}(end);
%dyn{j} = dyn{j}(end);

%% logdet
% extracting last element of the list (the diag values)
% tip from Matt jacobson@xorantech.com
%feat = cell2mat(dyn); %cellfun(@(x) x(end), dyn);
feat = [];

for k=1:fm-1
    load([PATH 'D' num2str(k) '.mat']);
    feat = [feat; cellfun(@(x) x(end), dyn)];
end

feat = [feat; featend];

% computing the logdet
res = 2 * sum(log(feat));
